import sys

import numpy as np
import pandas as pd
from inaSpeechSegmenter import Segmenter
from marblenet import marblenet
from rvad import rvad_fast
from soundfile import read, write
from tqdm import tqdm
from utils import gen_noise_signal, load_answer

import pysaten


def _main():
    rand = np.random.default_rng(0)

    wav_files = [
        "wav_and_lab/metan/ITA_recitation_normal_synchronized_wav/recitation",
        "wav_and_lab/sora/ITA_recitation_normal_synchronized_wav/recitation",
        "wav_and_lab/usagi/ITA_recitation_normal_synchronized_wav/normal_recitation_",
    ]

    lab_files = [
        "wav_and_lab/metan/ITA_recitation_normal_label/recitation",
        "wav_and_lab/sora/ITA_recitation_normal_label/rct",
        "wav_and_lab/usagi/ITA_recitation_normal_label/rct",
    ]

    marble = marblenet.MarbleNet()
    for snr in [None, 20, 15, 10, 5, 0, -5, -999]:
        saten = []
        rvad = []
        ina = []
        nemo = []
        print(f"SNR: {snr}", file=sys.stderr)
        for i in tqdm(range(1, 324 + 1)):
            for speaker in [0, 1, 2]:
                ans_s, ans_e = load_answer(f"{lab_files[speaker]}{i:03}.lab")
                wav_file = f"{wav_files[speaker]}{i:03}.wav"
                x, fs = read(wav_file)
                if snr is not None and snr != -999:
                    x = gen_noise_signal(x, fs, snr, True, rand, ans_s, ans_e)
                elif snr == -999:
                    x = rand.uniform(low=-1.0, high=1.0, size=len(x))
                # SATEN
                S, E = pysaten.vsed(x, fs)
                E = 0 if S == 0 and abs(E - len(x) / fs) <= 0.01 else E
                saten.append(abs(S - ans_s))
                saten.append(abs(E - ans_e))

                # rVAD
                S, E = rvad_fast.vad(x, fs)
                rvad.append(abs(S - ans_s))
                rvad.append(abs(E - ans_e))

                # ina
                S, E = _ina_speech_segmenter(x, fs)
                ina.append(abs(S - ans_s))
                ina.append(abs(E - ans_e))

                # nemo
                S, E = marble.vad_test(x, fs)
                nemo.append(abs(S - ans_s))
                nemo.append(abs(E - ans_e))

        pd.DataFrame(
            {"pySATEN": saten, "rVAD": rvad, "inaSpeechSegmenter": ina, "MarbleNet": nemo}
        ).to_csv(f"test_{str(snr)}.csv", index=False)


def _ina_speech_segmenter(x, fs):
    inaSegmenter = Segmenter(detect_gender=False)
    write("temp.wav", x, fs)
    segments = inaSegmenter("temp.wav")
    ina_temp = []
    for segment in segments:
        label, s, e = segment
        if label == "speech":
            ina_temp.append(s)
            ina_temp.append(e)
    S = ina_temp[0] if len(ina_temp) != 0 else 0
    E = ina_temp[-1] if len(ina_temp) != 0 else 0
    return S, E


if __name__ == "__main__":
    _main()



# SPDX-License-Identifier:GPL-3.0-or-later

import argparse
from typing import Optional

import noisereduce as nr
import numpy as np
import soundfile
from librosa import resample
from scipy.signal import cheby1, firwin, lfilter, sosfilt

from ._constant import F0_CEIL, F0_FLOOR, NYQ, SR
from ._signal import blue_noise, rms, zcr
from ._utility import normalize, slide_index


def cli_runner():
    # parse argument
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str)
    parser.add_argument("output", type=str)
    args = parser.parse_args()
    # trimming
    y, sr = soundfile.read(args.input)
    y_trimmed = trim(y, sr)
    soundfile.write(args.output, y_trimmed, sr)


def trim(y: np.ndarray, sr: int) -> np.ndarray:
    s_sec, e_sec = vsed(y, sr)
    return y[int(s_sec * sr) : int(e_sec * sr)]


def vsed(y: np.ndarray, sr: int) -> tuple[float, float]:
    _, _, _, _, start_s, end_s, _, _, _ = vsed_debug(y, sr)
    return start_s, end_s


def vsed_debug(
    y: np.ndarray,
    orig_sr: int,
    # -------------------------------------------
    win_length_s: Optional[float] = None,
    hop_length_s: float = 0.01,
    # -------------------------------------------
    rms_threshold: float = 0.03,
    zcr_threshold: float = 0.67,
    zcr_margin_s: float = 0.1,
    offset_s: float = 0.03,
    # -------------------------------------------
    noise_seed: int = 0,
):
    # resample
    y_rsp: np.ndarray = resample(
        y=y, orig_sr=orig_sr, target_sr=SR, res_type="soxr_lq"
    )

    # constants
    win_length_s = (
        win_length_s if win_length_s is not None else hop_length_s * 4
    )
    win_length: int = int(win_length_s * SR)
    hop_length: int = int(hop_length_s * SR)
    zcr_margin: int = int(zcr_margin_s / hop_length_s)

    # preprocess: add blue noise && remove background noise
    y_nr = _00_preprocess(y_rsp, SR, noise_seed)

    # step1: Root mean square
    start1, end1, y_rms = _01_rms(
        y_nr, SR, rms_threshold, win_length, hop_length
    )

    # step2: Zero cross
    start2, end2, y_zcr = _02_zcr(
        y_nr,
        SR,
        start1,
        end1,
        zcr_threshold,
        zcr_margin,
        win_length,
        hop_length,
    )

    # index -> second: rms
    start1_s: float = max(0, start1 * hop_length_s)
    end1_s: float = min(end1 * hop_length_s, len(y_rsp) / SR)

    # index -> second: zrs
    start2_s: float = max(0, start2 * hop_length_s)
    end2_s: float = min(end2 * hop_length_s, len(y_rsp) / SR)

    feats_timestamp = np.linspace(0, len(y_zcr) * hop_length_s, len(y_zcr))

    return (
        start1_s,
        end1_s,
        start2_s,
        end2_s,
        start2_s - offset_s,
        end2_s + offset_s,
        feats_timestamp,
        y_rms,
        y_zcr,
    )


def _00_preprocess(y: np.ndarray, sr: int, noise_seed: int) -> np.ndarray:
    data_rms = np.sort(rms(y, 2048, 512))
    signal_amp = data_rms[-2]
    noise_amp = max(data_rms[1], 1e-10)
    snr = min(20 * np.log10(signal_amp / noise_amp), 10)
    noise = blue_noise(len(y), sr, noise_seed)
    y_blue = y + noise * (signal_amp / 10 ** (snr / 20))
    y_blue = y_blue if max(abs(y_blue)) <= 1 else y_blue / max(abs(y_blue))
    return nr.reduce_noise(y_blue, sr)


def _01_rms(
    y, sr, threshold, win_length, hop_length
) -> tuple[int, int, np.ndarray]:
    wp = [F0_FLOOR / NYQ, F0_CEIL / NYQ]
    band_sos = cheby1(N=12, rp=1, Wn=wp, btype="bandpass", output="sos")
    y_bpf = sosfilt(band_sos, y)
    y_rms = normalize(rms(y_bpf, win_length, hop_length))
    start1: int = (
        np.where(threshold < y_rms)[0][0] if np.any(threshold < y_rms) else 0
    )
    end1: int = (
        np.where(threshold < y_rms)[0][-1]
        if np.any(threshold < y_rms)
        else len(y_rms) - 1
    )
    return start1, end1, y_rms


def _02_zcr(y, sr, start1, end1, threshold, margin, win_length, hop_length):
    high_b = firwin(101, F0_CEIL, pass_zero=False, fs=sr)
    y_hpf = lfilter(high_b, 1.0, y)
    y_zcr = normalize(zcr(y_hpf, win_length, hop_length))
    # slide start index
    start2 = slide_index(
        goto_min=True,
        y=y_zcr,
        start_idx=start1,
        threshold=threshold,
        margin=margin,
    )
    # slide end index
    end2 = slide_index(
        goto_min=False,
        y=y_zcr,
        start_idx=end1,
        threshold=threshold,
        margin=margin,
    )
    return start2, end2, y_zcr



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.ticker import MaxNLocator

snrlist = [None, 20, 15, 10, 5, 0, -5, -999]
vad = ["pySATEN", "rVAD", "inaSpeechSegmenter", "MarbleNet"]
result: dict = {
    "label": ["Inf", "20", "15", "10", "5", "0", "-5", "-Inf"],
    vad[0]: {"low": [], "med": [], "high": [], "width": []},
    vad[1]: {"low": [], "med": [], "high": [], "width": []},
    vad[2]: {"low": [], "med": [], "high": [], "width": []},
    vad[3]: {"low": [], "med": [], "high": [], "width": []},
}

print(f"SNR, {vad[0]}, {vad[1]}, {vad[2]}, {vad[3]}")
for snr in snrlist:
    trueSNR = (
        (snr if str(snr) != str(None) else "Inf") if str(snr) != str(-999) else "-Inf"
    )
    print(
        f"{trueSNR}, ",
        end="",
    )
    file_path = f"test_{str(snr)}.csv"
    df = pd.read_csv(file_path)
    for method in vad:
        result[method]["low"].append(df[method].quantile(0.25))
        result[method]["med"].append(df[method].median())
        result[method]["high"].append(df[method].quantile(0.75))
        result[method]["width"].append(
            df[method].quantile(0.75) - df[method].quantile(0.25)
        )
        print(f"{float(result[method]['med'][-1]):.3f} ", end="")
        print(
            f"({float(result[method]['low'][-1]):.3f}"
            + f":{float(result[method]['high'][-1]):.3f}"
            + f": {result[method]['width'][-1]:.3f}), ",
            end="",
        )
    print("")

# color: https://contents-open.hatenablog.com/entry/2021/08/19/231157
plt.rcParams["font.family"] = "Times New Roman"
plt.rcParams["font.size"] = 16
x = np.arange(len(result["label"]))
offset = 0.11
markersize = 7
capsize = 0
linewidth = 2
plt.errorbar(
    x + 1.5 * offset,
    result[vad[3]]["med"],
    yerr=[result[vad[3]]["low"], result[vad[3]]["high"]],
    fmt="*:",
    label=str(vad[3]),
    capsize=capsize,
    markersize=markersize * 1.5,
    linewidth=linewidth,
    color="#FF4B00",
)
plt.errorbar(
    x + 0.5 * offset,
    result[vad[2]]["med"],
    yerr=[result[vad[2]]["low"], result[vad[2]]["high"]],
    fmt="D-.",
    label=str(vad[2]),
    capsize=capsize,
    markersize=markersize,
    linewidth=linewidth,
    color="#005AFF",
)
plt.errorbar(
    x - 0.5 * offset,
    result[vad[1]]["med"],
    yerr=[result[vad[1]]["low"], result[vad[1]]["high"]],
    fmt="s--",
    label=str(vad[1]),
    capsize=capsize,
    markersize=markersize,
    linewidth=linewidth,
    color="#03AF7A",
)
plt.errorbar(
    x - 1.5 * offset,
    result[vad[0]]["med"],
    yerr=[result[vad[0]]["low"], result[vad[0]]["high"]],
    fmt="o-",
    label="SATEN Lv.1",
    capsize=capsize,
    markersize=markersize,
    linewidth=linewidth,
    color="#4DC4FF",
)
plt.xticks(x, result["label"])
plt.legend(handlelength=4, fontsize=12, loc="upper left")
# plt.ylim(-0.5, 3.5)
plt.gca().yaxis.set_major_locator(MaxNLocator(integer=True))
plt.grid()
plt.yscale("log")
plt.ylabel("Error [s]", fontsize=20)
plt.xlabel("Signal-to-Noise Ratio [dB]", fontsize=20)
plt.show()



import numpy as np


def rms(y, win_length, hop_length):
    rms = np.zeros(int(np.ceil(len(y) / hop_length)))
    for i in range(len(rms)):
        # get target array
        idx = i * hop_length
        zc_start = int(max(0, idx - (win_length / 2)))
        zc_end = int(min(idx + (win_length / 2), len(y) - 1))
        target = y[zc_start:zc_end]
        # calc rms
        rms[i] = np.sqrt(np.mean(pow(target, 2)))
    return rms


def zcr(y, win_length, hop_length):
    zc = np.zeros(int(np.ceil(len(y) / hop_length)))
    for i in range(len(zc)):
        # get target array
        idx = i * hop_length
        zc_start = int(max(0, idx - (win_length / 2)))
        zc_end = int(min(idx + (win_length / 2), len(y) - 1))
        target = y[zc_start:zc_end]
        # calc zcr
        sign_arr = np.sign(target)[target != 0 & ~np.isnan(target)]
        zc[i] = np.sum(np.abs(np.diff(sign_arr)) != 0) / hop_length
    return zc


def blue_noise(length: int, sr: int, noise_seed: int) -> np.ndarray:
    rand: np.random.Generator = np.random.default_rng(noise_seed)
    length2 = int(pow(2, np.ceil(np.log2(length)) + 1))
    # white noise
    wh = rand.uniform(low=-1.0, high=1.0, size=length2)
    # fft
    WH = np.fft.rfft(wh)
    WH_f = np.fft.rfftfreq(len(wh), 1 / sr)
    # white -> blue
    BL = WH * np.sqrt(WH_f)
    # irfft
    bl = np.fft.irfft(BL)
    # normalize
    bl /= np.max(np.abs(bl))

    return bl[:length]



from __future__ import division
import numpy as np
from scipy.signal import lfilter
import rvad.speechproc
from copy import deepcopy


# Refs:
#  [1] Z.-H. Tan, A.k. Sarkara and N. Dehak, "rVAD: an unsupervised segment-based robust voice activity detection method," Computer Speech and Language, vol. 59, pp. 1-21, 2020. 
#  [2] Z.-H. Tan and B. Lindberg, "Low-complexity variable frame rate analysis for speech recognition and voice activity detection." 
#  IEEE Journal of Selected Topics in Signal Processing, vol. 4, no. 5, pp. 798-807, 2010.

# Version: 2.0
# 02 Dec 2017, Achintya Kumar Sarkar and Zheng-Hua Tan

def vad(data, fs):
    winlen, ovrlen, pre_coef, nfilter, nftt = 0.025, 0.01, 0.97, 20, 512
    ftThres = 0.5; vadThres = 0.4
    opts = 1

    ft, flen, fsh10, nfr10 = rvad.speechproc.sflux(
        data, fs, winlen, ovrlen, nftt)

    # --spectral flatness --
    pv01 = np.zeros(nfr10)
    pv01[np.less_equal(ft, ftThres)] = 1 
    pitch = deepcopy(ft)

    pvblk = rvad.speechproc.pitchblockdetect(
        pv01, pitch, nfr10, opts)


    # --filtering--
    ENERGYFLOOR = np.exp(-50)
    b = np.array([0.9770, -0.9770])
    a = np.array([1.0000, -0.9540])
    fdata=lfilter(b, a, data, axis=0)


    #--pass 1--
    noise_samp, noise_seg, n_noise_samp = rvad.speechproc.snre_highenergy(
        fdata, nfr10, flen, fsh10, ENERGYFLOOR, pv01, pvblk)

    #sets noisy segments to zero
    for j in range(n_noise_samp):
        fdata[range(int(noise_samp[j,0]), int(noise_samp[j,1]) +1)] = 0 


    vad_seg = rvad.speechproc.snre_vad(
        fdata, nfr10, flen, fsh10, ENERGYFLOOR, pv01, pvblk, vadThres)


    start = np.min(np.where(vad_seg != 0)) * ovrlen if np.any(vad_seg != 0) else 0
    end = np.max(np.where(vad_seg != 0)) * ovrlen if np.any(vad_seg != 0) else 0

    return start, end



import numpy as np


def slide_index(
    goto_min: bool,
    y: np.ndarray,
    start_idx: int,
    threshold: float,
    margin: int,
) -> int:

    stop_idx: int = -1 if goto_min else len(y)
    step: int = -1 if goto_min else 1

    for i in range(start_idx, stop_idx, step):
        if threshold <= y[i]:
            a_check_end = (
                max(0, i - margin) if goto_min else min(i + margin, len(y))
            )
            a_check = y[a_check_end:i] if goto_min else y[i:a_check_end]
            indices_below_threshold = [
                j for j, b in enumerate(a_check) if b < threshold
            ]
            if indices_below_threshold:  # is not empty
                i = (
                    min(indices_below_threshold)
                    if goto_min
                    else max(indices_below_threshold)
                )
            else:  # indices_below_threshold is empty -> finish!!!
                return i
    return 0


def normalize(y: np.ndarray) -> np.ndarray:
    return (y - min(y)) / (max(y) - min(y))
